# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate project to the latest Renku version."""

from typing import List

from pydantic import validate_arguments

from renku.command.command_builder.command import Command
from renku.domain_model.project_context import project_context

SUPPORTED_RENKU_PROJECT = 1
MIGRATION_REQUIRED = 2
UNSUPPORTED_PROJECT = 4
NON_RENKU_REPOSITORY = 8
TEMPLATE_UPDATE_POSSIBLE = 16
AUTOMATED_TEMPLATE_UPDATE_SUPPORTED = 32
DOCKERFILE_UPDATE_POSSIBLE = 64


def migrations_check():
    """Return a command for a migrations check."""
    return Command().command(_migrations_check).with_database(write=False)


def _migrations_check():
    """Check migration status of project.

    Returns:
        Dictionary of project migrations, template and dockerfile status.
    """
    from renku.core.migration.migrate import is_project_unsupported

    core_version, latest_version = _migrations_versions()

    return {
        "project_supported": not is_project_unsupported(),
        "core_renku_version": core_version,
        "project_renku_version": latest_version,
        "core_compatibility_status": _metadata_migration_check(),
        "dockerfile_renku_status": _dockerfile_migration_check(),
        "template_status": _template_migration_check(),
    }


def migrations_versions():
    """Return a command to get source and destination migration versions."""
    return Command().command(_migrations_versions).lock_project().with_database()


def _migrations_versions():
    """Return source and destination migration versions.

    Returns:
        Tuple of current version and project version.
    """
    from renku import __version__

    try:
        latest_agent = project_context.latest_agent
    except ValueError:
        # NOTE: maybe old project
        from renku.core.migration.utils import read_latest_agent

        latest_agent = read_latest_agent()

    return __version__, latest_agent


def _template_migration_check():
    """Return template migration status.

    Returns:
        Dictionary of template migration status.
    """
    from renku.core.template.usecase import check_for_template_update

    try:
        project = project_context.project
        template_source = project.template_metadata.template_source
        template_ref = project.template_metadata.template_ref
        template_id = project.template_metadata.template_id
    except (ValueError, AttributeError):
        project = None
        template_source = None
        template_ref = None
        template_id = None

    update_available, update_allowed, current_version, new_version = check_for_template_update(project)

    return {
        "automated_template_update": update_allowed,
        "newer_template_available": update_available,
        "project_template_version": current_version,
        "latest_template_version": new_version,
        "template_source": template_source,
        "template_ref": template_ref,
        "template_id": template_id,
    }


def dockerfile_migration_check():
    """Return a command for a Dockerfile migrations check."""
    return Command().command(_dockerfile_migration_check)


def _dockerfile_migration_check():
    """Return Dockerfile migration status.

    Returns:
        Dictionary of Dockerfile migration status.
    """
    from renku import __version__
    from renku.core.migration.migrate import is_docker_update_possible

    automated_dockerfile_update, newer_renku_available, dockerfile_renku_version = is_docker_update_possible()

    return {
        "automated_dockerfile_update": automated_dockerfile_update,
        "newer_renku_available": newer_renku_available,
        "dockerfile_renku_version": dockerfile_renku_version,
        "latest_renku_version": __version__,
    }


def metadata_migration_check():
    """Return a command for a metadata migrations check."""
    return Command().command(_metadata_migration_check)


def _metadata_migration_check():
    """Return metadata migration status.

    Returns:
        Dictionary of metadata migration status.
    """
    from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION, get_project_version, is_migration_required

    return {
        "migration_required": is_migration_required(),
        "project_metadata_version": get_project_version(),
        "current_metadata_version": SUPPORTED_PROJECT_VERSION,
    }


def migrate_project_command():
    """Return a command to migrate all project's entities."""
    from renku.core.migration.migrate import migrate_project

    return Command().command(migrate_project).lock_project().require_clean().with_database(write=True)


def check_project():
    """Return a command to check if repository is a renku project, unsupported, or requires migration."""
    return Command().command(_check_project).with_database(write=False)


def _check_project():
    from renku.core.migration.migrate import (
        is_docker_update_possible,
        is_migration_required,
        is_project_unsupported,
        is_renku_project,
    )
    from renku.core.template.usecase import check_for_template_update

    if not is_renku_project():
        return NON_RENKU_REPOSITORY
    elif is_project_unsupported():
        return UNSUPPORTED_PROJECT

    try:
        _ = project_context.project
    except ValueError:
        return MIGRATION_REQUIRED
    else:
        if hasattr(project_context.project, "template_source"):
            # NOTE: v10 migration not done
            return MIGRATION_REQUIRED

    # NOTE: ``project.automated_update`` is deprecated and we always allow template update for a project
    status = AUTOMATED_TEMPLATE_UPDATE_SUPPORTED

    if check_for_template_update(project_context.project)[0]:
        status |= TEMPLATE_UPDATE_POSSIBLE
    if is_docker_update_possible()[0]:
        status |= DOCKERFILE_UPDATE_POSSIBLE

    if is_migration_required():
        return status | MIGRATION_REQUIRED

    return status | SUPPORTED_RENKU_PROJECT


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _check_immutable_template_files(paths: List[str]):
    """Check paths and return a list of those that are marked immutable in the project template.

    Args:
        paths(List[str]): Paths to check.

    Returns:
        List of immutable template files.
    """
    immutable_template_files = project_context.project.template_metadata.immutable_template_files or []

    return [p for p in paths if str(p) in immutable_template_files]


def check_immutable_template_files_command():
    """Command for checking immutable template files."""
    return Command().command(_check_immutable_template_files).with_database()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
from flask import request

from renku.ui.service.controllers.cache_migrate_project import MigrateProjectCtrl
from renku.ui.service.serializers.v1_0.cache import ProjectMigrateResponseRPC_1_0
from renku.ui.service.views.api_versions import V0_9, V1_0
from renku.ui.service.views.decorators import accepts_json, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import handle_common_except, handle_migration_write_errors


@handle_common_except
@handle_migration_write_errors
@accepts_json
@requires_cache
@requires_identity
def migrate_project_view_1_0(user_data, cache):
    """
    Migrate a project.

    ---
    post:
      description: Migrate a project.
      requestBody:
        content:
          application/json:
            schema: ProjectMigrateRequest
      responses:
        200:
          description: Migration status.
          content:
            application/json:
              schema: ProjectMigrateResponseRPC_1_0
      tags:
        - cache
    """
    ctrl = MigrateProjectCtrl(cache, user_data, dict(request.json))  # type: ignore
    ctrl.RESPONSE_SERIALIZER = ProjectMigrateResponseRPC_1_0()  # type: ignore
    return ctrl.to_response()


def add_v1_0_specific_endpoints(cache_blueprint):
    """Add v1.0 only endpoints to blueprint."""
    cache_blueprint.route("/cache.migrate", methods=["POST"], provide_automatic_options=False, versions=[V0_9, V1_0])(
        migrate_project_view_1_0
    )
    return cache_blueprint



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General utility functions."""

from typing import Any, NewType, Optional

from packaging.version import Version


def to_string(value: Optional[Any], strip: bool = False) -> str:
    """Return a string representation of value and return an empty string for None."""
    return "" if value is None else str(value).strip() if strip else str(value)


def to_semantic_version(value: str) -> Optional[Version]:
    """Convert value to SemVer."""
    try:
        return Version(value)
    except ValueError:
        return None


NoValueType = NewType("NoValueType", object)
"""Type to represent a value not being set in cases where ``None`` is a valid value."""

NO_VALUE = NoValueType(object())
"""Sentinel to represent a value not being set in cases where ``None`` is a valid value."""



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""S3 dataset provider."""

import re
import urllib
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ProviderApi,
    ProviderCredentials,
    ProviderPriority,
    StorageProviderInterface,
)
from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata, ProviderParameter
from renku.core.interface.storage import IStorage, IStorageFactory
from renku.core.util.metadata import prompt_for_credentials
from renku.core.util.urls import get_scheme
from renku.domain_model.dataset import RemoteEntity
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.domain_model.dataset import Dataset


class S3Provider(ProviderApi, AddProviderInterface, StorageProviderInterface):
    """S3 provider."""

    priority = ProviderPriority.HIGHEST
    name = "S3"

    def __init__(self, uri: Optional[str]):
        super().__init__(uri=uri)

        endpoint, bucket, _ = parse_s3_uri(uri=self.uri)

        self._bucket: str = bucket
        self._endpoint: str = endpoint

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        return get_scheme(uri) == "s3"

    @staticmethod
    def get_add_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for add."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter(
                "storage",
                flags=["storage"],
                default=None,
                help="Uri for the S3 bucket when creating the dataset at the same time when running 'add'",
                multiple=False,
                type=str,
            ),
        ]

    @inject.autoparams("storage_factory")
    def get_storage(
        self, storage_factory: "IStorageFactory", credentials: Optional["ProviderCredentials"] = None
    ) -> "IStorage":
        """Return the storage manager for the provider."""
        s3_configuration = {
            "type": "s3",
            "provider": "AWS",
            "endpoint": self.endpoint,
        }

        def create_renku_storage_s3_uri(uri: str) -> str:
            """Create a S3 URI to work with the Renku storage handler."""
            _, bucket, path = parse_s3_uri(uri=uri)

            return f"s3://{bucket}/{path}"

        if not credentials:
            credentials = S3Credentials(provider=self)
            prompt_for_credentials(credentials)

        return storage_factory.get_storage(
            storage_scheme="s3",
            provider=self,
            credentials=credentials,
            configuration=s3_configuration,
            uri_convertor=create_renku_storage_s3_uri,
        )

    def add(self, uri: str, destination: Path, **kwargs) -> List["DatasetAddMetadata"]:
        """Add files from a URI to a dataset."""
        if re.search(r"[*?]", uri):
            raise errors.ParameterError("Wildcards like '*' or '?' are not supported for S3 URIs.")

        storage = self.get_storage()

        destination_path_in_repo = Path(destination).relative_to(project_context.repository.path)
        hashes = storage.get_hashes(uri=uri)
        return [
            DatasetAddMetadata(
                entity_path=destination_path_in_repo / hash.path,
                url=hash.base_uri,
                action=DatasetAddAction.REMOTE_STORAGE,
                # TODO: Store the original URI for use as source; based_on is not needed
                based_on=RemoteEntity(checksum=hash.hash if hash.hash else "", url=hash.base_uri, path=hash.path),
                source=Path(hash.base_uri),
                destination=destination_path_in_repo,
                provider=self,
            )
            for hash in hashes
        ]

    @property
    def bucket(self) -> str:
        """Return S3 bucket name."""
        return self._bucket

    @property
    def endpoint(self) -> str:
        """Return S3 bucket endpoint."""
        return self._endpoint

    def on_create(self, dataset: "Dataset") -> None:
        """Hook to perform provider-specific actions on a newly-created dataset."""
        credentials = S3Credentials(provider=self)
        prompt_for_credentials(credentials)
        storage = self.get_storage(credentials=credentials)

        # NOTE: The underlying rclone tool cannot tell if a directory within a S3 bucket exists or not
        if not storage.exists(self.uri):
            raise errors.ParameterError(f"S3 bucket '{self.bucket}' doesn't exists.")

        project_context.repository.add_ignored_pattern(pattern=str(dataset.get_datadir()))


class S3Credentials(ProviderCredentials):
    """S3-specific credentials."""

    def __init__(self, provider: S3Provider):
        super().__init__(provider=provider)

    @staticmethod
    def get_credentials_names() -> Tuple[str, ...]:
        """Return a tuple of the required credentials for a provider."""
        return "Access Key ID", "Secret Access Key"

    @property
    def provider(self) -> S3Provider:
        """Return the associated provider instance."""
        return cast(S3Provider, self._provider)

    def get_credentials_section_name(self) -> str:
        """Get section name for storing credentials.

        NOTE: This methods should be overridden by subclasses to allow multiple credentials per providers if needed.
        """
        return self.provider.endpoint.lower()


def parse_s3_uri(uri: str) -> Tuple[str, str, str]:
    """Extract endpoint, bucket name, and path within the bucket from a given URI.

    NOTE: We only support s3://<endpoint>/<bucket-name>/<path> at the moment.
    """
    parsed_uri = urllib.parse.urlparse(uri)

    endpoint = parsed_uri.netloc
    path = parsed_uri.path.strip("/")

    if parsed_uri.scheme.lower() != "s3" or not endpoint:
        raise errors.ParameterError(f"Invalid S3 URI: {uri}. Valid format is 's3://<endpoint>/<bucket-name>/<path>'")

    bucket, _, path = path.partition("/")

    return endpoint, bucket, path.strip("/")



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template variable utility methods."""

import datetime
from string import Formatter
from typing import Any, Iterable, Mapping, Tuple, Union

from renku.core.errors import ParameterError
from renku.domain_model.workflow.parameter import CommandParameterBase


class TemplateVariableFormatter(Formatter):
    """Template variable formatter for `CommandParameterBase`."""

    RESERVED_KEYS = ["iter_index"]

    def __init__(self):
        super(TemplateVariableFormatter, self).__init__()

    def apply(self, param: str, parameters: Mapping[str, Any] = {}) -> str:
        """Renders the parameter template into its final value."""
        try:
            return super().vformat(param, args=[datetime.datetime.now()], kwargs=parameters)
        except KeyError as e:
            raise ParameterError(f"Could not resolve the variable {str(e)}")

    def get_value(self, key, args, kwargs):
        """Ignore some special keys when formatting the variable."""
        if key in self.RESERVED_KEYS:
            return key
        return super().get_value(key, args, kwargs)

    @staticmethod
    def to_map(parameters: Iterable[Union[CommandParameterBase, Tuple[str, str]]]) -> Mapping[str, str]:
        """Converts a list of `CommandParameterBase` into parameter name-value dictionary."""
        return dict(
            map(
                lambda x: (x.name, x.actual_value) if isinstance(x, CommandParameterBase) else (x[1], str(x[0])),
                parameters,
            )
        )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pointer file business logic."""

import os
import uuid
from pathlib import Path
from typing import Optional, Tuple, Union, cast

from renku.core import errors
from renku.core.util.os import is_subpath
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Repository


def create_pointer_file(target: Union[str, Path], checksum: str = None):
    """Create a new pointer file."""
    target = Path(target).resolve()

    if checksum is None:
        checksum = Repository.hash_object(target)
        assert checksum is not None, f"Cannot calculate checksum for '{target}'"

    while True:
        filename = f"{uuid.uuid4()}-{checksum}"
        path = project_context.pointers_path / filename
        if not path.exists():
            break

    # NOTE: If target is within the repo, add it as a relative symlink
    is_within_repo = is_subpath(target, base=project_context.path)
    source = cast(Union[str, bytes, Path], os.path.relpath(target, path.parent) if is_within_repo else target)

    try:
        os.symlink(source, path)
    except FileNotFoundError:
        raise errors.ExternalFileNotFound(target)

    return path


def is_external_file_updated(project_path: Path, path: Union[Path, str]) -> Tuple[bool, str]:
    """Check if an update to an external file is available."""
    pointer_file = get_pointer_file(project_path=project_path, path=path)

    try:
        target = pointer_file.resolve(strict=True)
    except FileNotFoundError:
        target = pointer_file.resolve()
        raise errors.ExternalFileNotFound(target)

    new_checksum = Repository.hash_object(target)
    old_checksum = pointer_file.name.split("-")[-1]

    if new_checksum is None:
        raise errors.ExternalFileNotFound(target)

    updated = new_checksum != old_checksum

    return updated, new_checksum


def update_external_file(path: Union[Path, str], checksum: Optional[str]):
    """Delete existing external file and create a new one."""
    pointer_file = get_pointer_file(project_path=project_context.path, path=path)
    target = pointer_file.resolve()

    os.remove(pointer_file)
    absolute_path = project_context.path / path
    os.remove(absolute_path)

    create_external_file(target=target, path=absolute_path, checksum=checksum)


def create_external_file(target: Path, path: Union[Path, str], checksum: str = None):
    """Create a new external file."""
    try:
        pointer_file = create_pointer_file(target=target, checksum=checksum)
        relative = os.path.relpath(pointer_file, Path(path).parent)
        os.symlink(relative, path)
    except OSError as e:
        raise errors.OperationError("Could not create symbolic link") from e


def get_pointer_file(project_path: Path, path: Union[Path, str]) -> Path:
    """Return pointer file from an external file."""
    absolute_path = project_path / path
    link = absolute_path.parent / os.readlink(absolute_path)
    return project_path / link



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
from flask import request

from renku.ui.service.controllers.cache_migrations_check import MigrationsCheckCtrl
from renku.ui.service.gateways.gitlab_api_provider import GitlabAPIProvider
from renku.ui.service.serializers.v0_9.cache import ProjectMigrationCheckResponseRPC_0_9
from renku.ui.service.views.api_versions import V0_9
from renku.ui.service.views.decorators import optional_identity, requires_cache
from renku.ui.service.views.error_handlers import handle_common_except


@handle_common_except
@requires_cache
@optional_identity
def migration_check_project_view_0_9(user_data, cache):
    """
    Retrieve migration information for a project.

    ---
    get:
      description: Retrieve migration information for a project.
      parameters:
        - in: query
          schema: ProjectMigrationCheckRequest
      responses:
        200:
          description: Information about required migrations for the project.
          content:
            application/json:
              schema: ProjectMigrationCheckResponseRPC_0_9
      tags:
        - cache
    """
    ctrl = MigrationsCheckCtrl(cache, user_data, dict(request.args), GitlabAPIProvider())
    ctrl.RESPONSE_SERIALIZER = ProjectMigrationCheckResponseRPC_0_9()  # type: ignore
    return ctrl.to_response()


def add_v0_9_specific_endpoints(cache_blueprint):
    """Add v0.9 only endpoints to blueprint."""
    cache_blueprint.route("/cache.migrations_check", methods=["GET"], provide_automatic_options=False, versions=[V0_9])(
        migration_check_project_view_0_9
    )
    return cache_blueprint



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DOI API integration."""

import urllib
from pathlib import Path
from typing import Optional

from renku.core import errors
from renku.core.dataset.providers.api import ImporterApi, ImportProviderInterface, ProviderApi, ProviderPriority
from renku.core.util.doi import extract_doi, is_doi

DOI_BASE_URL = "https://dx.doi.org"


class DOIProvider(ProviderApi, ImportProviderInterface):
    """`doi.org <http://doi.org>`_ registry API provider."""

    priority = ProviderPriority.HIGHER
    name = "DOI"

    def __init__(self, uri: Optional[str], headers=None, timeout=3):
        super().__init__(uri=uri)

        self.timeout = timeout
        self.headers = headers if headers is not None else {"accept": "application/vnd.citationstyles.csl+json"}

    @staticmethod
    def supports(uri) -> bool:
        """Whether or not this provider supports a given URI."""
        return bool(is_doi(uri))

    def get_importer(self, **kwargs) -> "DOIImporter":
        """Get import manager."""
        from renku.core.util import requests

        def query(doi):
            """Retrieve metadata for given doi."""
            doi = extract_doi(doi)
            url = make_doi_url(doi)

            response = requests.get(url, headers=self.headers)

            if response.status_code != 200:
                raise LookupError("record not found. Status: {}".format(response.status_code))

            return response

        def serialize(response):
            """Serialize HTTP response for DOI."""
            json_data = response.json()
            data = {key.replace("-", "_").lower(): value for key, value in json_data.items()}
            try:
                return DOIImporter(**data)
            except TypeError:
                raise errors.DatasetImportError("doi metadata could not be serialized")

        query_response = query(self.uri)
        return serialize(query_response)


class DOIImporter(ImporterApi):
    """Response from `doi.org <http://doi.org>`_ for DOI metadata."""

    def __init__(
        self,
        id,
        doi,
        url,
        abstract=None,
        author=None,
        categories=None,
        container_title=None,
        contributor=None,
        copyright=None,
        issued=None,
        language=None,
        publisher=None,
        title=None,
        type=None,
        version=None,
    ):
        super().__init__(uri=url, original_uri=url)

        self.id = id
        self.doi = doi

        self.abstract = abstract
        self.author = author
        self.categories = categories
        self.container_title = container_title
        self.contributor = contributor
        self.copyright = copyright
        self.issued = issued
        self.language = language
        self.publisher = publisher
        self.title = title
        self.type = type
        self._version = version

    @property
    def version(self) -> str:
        """Get record version."""
        return self._version

    @property
    def latest_uri(self) -> str:
        """Get URI of the latest version."""
        return self.uri

    def fetch_provider_dataset(self):
        """Deserialize this record to a ``ProviderDataset``."""
        raise NotImplementedError

    def is_latest_version(self) -> bool:
        """Check if record is at last possible version."""
        return True

    def download_files(self, destination: Path, extract: bool):
        """Download dataset files from the remote provider."""
        raise NotImplementedError

    def tag_dataset(self, name: str) -> None:
        """Create a tag for the dataset ``name`` if the remote dataset has a tag/version."""
        raise NotImplementedError

    def copy_extra_metadata(self, new_dataset) -> None:
        """Copy provider specific metadata once the dataset is created."""
        raise NotImplementedError


def make_doi_url(doi):
    """Create URL to access DOI metadata."""
    parsed_url = urllib.parse.urlparse(doi)
    if parsed_url.scheme == "doi":
        parsed_url = parsed_url._replace(scheme="")
        doi = parsed_url.geturl()
    return urllib.parse.urljoin(DOI_BASE_URL, doi)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset add business logic."""

import itertools
import os.path
import shutil
from pathlib import Path
from typing import Generator, List, Optional, Union, cast

from renku.core import errors
from renku.core.dataset.context import DatasetContext
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.dataset.pointer_file import create_external_file
from renku.core.dataset.providers.api import ImporterApi
from renku.core.dataset.providers.factory import ProviderFactory
from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata
from renku.core.interface.storage import IStorage
from renku.core.storage import check_external_storage, track_paths_in_storage
from renku.core.util import communication, requests
from renku.core.util.dataset import check_url
from renku.core.util.git import get_git_user
from renku.core.util.os import delete_dataset_file, get_files, get_relative_path, hash_file
from renku.core.util.urls import is_uri_subfolder
from renku.domain_model.dataset import Dataset, DatasetFile, RemoteEntity
from renku.domain_model.project_context import project_context


def add_to_dataset(
    dataset_name: str,
    urls: List[str],
    *,
    importer: Optional[ImporterApi] = None,
    force: bool = False,
    create: bool = False,
    overwrite: bool = False,
    sources: Optional[List[Union[str, Path]]] = None,
    destination: str = "",
    revision: Optional[str] = None,
    extract: bool = False,
    clear_files_before: bool = False,
    total_size: Optional[int] = None,
    datadir: Optional[Path] = None,
    storage: Optional[str] = None,
    **kwargs,
) -> Dataset:
    """Import the data into the data directory."""
    sources = sources or []

    if not create and storage:
        raise errors.ParameterError("Storage can be set only when creating a dataset")

    try:
        with DatasetContext(name=dataset_name, create=create, datadir=datadir, storage=storage) as dataset:
            destination_path = _create_destination_directory(dataset, destination)

            check_external_storage()

            # NOTE: This is not required for external storages
            if not dataset.storage:
                _check_available_space(urls, total_size=total_size)

            datadir = cast(Path, project_context.path / dataset.get_datadir())
            if create and datadir.exists() and not dataset.storage:
                # NOTE: Add datadir to paths to add missing files on create
                for file in get_files(datadir):
                    urls.append(str(file))

            files = _get_files_metadata(
                urls=urls,
                dataset=dataset,
                importer=importer,
                destination=destination_path,
                revision=revision,
                sources=sources,
                extract=extract,
                force=force,
                **kwargs,
            )

            files = filter_files(dataset=dataset, files=files, force=force, overwrite=overwrite)
            if not files:
                if create:
                    raise errors.UsageError("There are no files to create a dataset")
                else:
                    communication.warn("No new file was added to project")
                    return dataset

            # NOTE: All files at this point can be force-added

            move_files_to_dataset(dataset=dataset, files=files)
            add_files_to_repository(dataset=dataset, files=files)
            update_dataset_metadata(dataset=dataset, files=files, clear_files_before=clear_files_before)

            DatasetsProvenance().add_or_update(dataset, creator=get_git_user(repository=project_context.repository))
    except errors.DatasetNotFound:
        raise errors.DatasetNotFound(
            message="Dataset '{0}' does not exist.\n"
            "Use 'renku dataset create {0}' to create the dataset or retry 'renku dataset add {0}' command "
            "with '--create' option for automatic dataset creation.".format(dataset_name)
        )
    except (FileNotFoundError, errors.GitCommandError) as e:
        raise errors.ParameterError("Could not find paths/URLs: \n{0}".format("\n".join(urls))) from e
    else:
        project_context.database.commit()
        return dataset


def _get_files_metadata(
    *,
    urls: List[str],
    importer: Optional[ImporterApi] = None,
    dataset: Dataset,
    destination: Path,
    extract: bool,
    revision: Optional[str],
    sources: List[Union[str, Path]],
    force: bool = False,
    **kwargs,
) -> List[DatasetAddMetadata]:
    """Process file URLs for adding to a dataset."""
    if importer:
        return importer.download_files(destination=destination, extract=extract)

    if len(urls) == 0:
        raise errors.ParameterError("No URL is specified")
    if sources and len(urls) > 1:
        raise errors.ParameterError("Cannot use '--source' with multiple URLs.")

    files = []

    for url in urls:
        _, is_git = check_url(url)

        if not is_git and sources:
            raise errors.ParameterError("Cannot use '-s/--src/--source' with URLs or local files.")

        provider = ProviderFactory.get_add_provider(uri=url)

        new_files = provider.add(
            uri=url,
            destination=destination,
            revision=revision,
            sources=sources,
            dataset=dataset,
            extract=extract,
            force=force,
            **kwargs,
        )

        files.extend(new_files)

    return files


def _check_available_space(urls: List[str], total_size: Optional[int] = None):
    """Check that there is enough space available on the device for download."""
    if total_size is None:
        total_size = 0
        for url in urls:
            try:
                response = requests.head(url, allow_redirects=True)
                total_size += int(response.headers.get("content-length", 0))
            except errors.RequestError:
                pass
    usage = shutil.disk_usage(project_context.path)

    if total_size > usage.free:
        mb = 2**20
        message = "Insufficient disk space (required: {:.2f} MB" "/available: {:.2f} MB). ".format(
            total_size / mb, usage.free / mb
        )
        raise errors.OperationError(message)


def _create_destination_directory(dataset: Dataset, destination: Optional[Union[Path, str]] = None) -> Path:
    """Create directory for dataset add."""
    dataset_datadir = project_context.path / dataset.get_datadir()

    if dataset_datadir.is_symlink():
        dataset_datadir.unlink()

    # NOTE: Make sure that dataset's data dir exists because we check for existence of a destination later to decide
    # what will be its name
    dataset_datadir.mkdir(parents=True, exist_ok=True)

    destination = destination or ""
    relative_path = cast(str, get_relative_path(destination, base=dataset_datadir, strict=True))
    return dataset_datadir / relative_path


def filter_files(
    dataset: Dataset, files: List[DatasetAddMetadata], force: bool, overwrite: bool
) -> List[DatasetAddMetadata]:
    """Filter ignored and overwritten files."""

    def remove_git_files(files_to_filter: List[DatasetAddMetadata]):
        """Remove all files that are under a .git directory."""
        git_paths = [f.entity_path for f in files_to_filter if str(f.entity_path).startswith(".git")]
        if not git_paths:
            return files_to_filter

        communication.warn("Ignored adding paths under a .git directory:\n\t" + "\n\t".join(str(p) for p in git_paths))
        return [f for f in files_to_filter if f.entity_path not in git_paths]

    def check_ignored_files(files_to_filter: Generator[DatasetAddMetadata, None, None]):
        """Check if any files added were ignored."""
        paths = {f.get_absolute_commit_path(project_context.path): f for f in files_to_filter}

        ignored_paths = project_context.repository.get_ignored_paths(*paths)
        if ignored_paths:
            ignored_sources = [file.source for path, file in paths.items() if path in ignored_paths]

            communication.warn(
                "Theses paths are ignored by one of your .gitignore files (use '--force' flag if you really want to "
                "add them):\n\t" + "\n\t".join([str(p) for p in ignored_sources])
            )

        return (file for path, file in paths.items() if path not in ignored_paths)

    def check_existing_files(files_to_filter: Generator[DatasetAddMetadata, None, None]):
        """Check if files added already exist."""
        files_list = list(files_to_filter)
        existing_paths = [f.entity_path for f in files_list if dataset.find_file(f.entity_path)]
        if existing_paths:
            communication.warn(
                "These existing files were not overwritten (use '--overwrite' flag to overwrite them):\n\t"
                + "\n\t".join([str(p) for p in existing_paths])
            )

        return (f for f in files_list if f.entity_path not in existing_paths)

    files = remove_git_files(files)

    # NOTE: Don't filter ignored or existing files that will be added to a remote storage
    remote_files = (f for f in files if f.metadata_only)
    local_files = (f for f in files if not f.metadata_only)

    # NOTE: Data directory of datasets with a storage backend is always ignored, so, filtering files is meaningless
    if not force and not dataset.storage:
        local_files = check_ignored_files(local_files)

    if not overwrite:
        local_files = check_existing_files(local_files)

    files = list(itertools.chain(local_files, remote_files))

    return files


def get_dataset_file_path_within_dataset(dataset: Dataset, entity_path: Union[Path, str]) -> Path:
    """Return a dataset file's path relative to the dataset's datadir.

    NOTE: Dataset files can reside outside the dataset's datadir.
    """
    assert not os.path.isabs(entity_path), f"Entity path cannot be absolute: {entity_path}"

    entity_path = Path(entity_path)

    try:
        return entity_path.relative_to(dataset.get_datadir())
    except ValueError:
        return entity_path


def get_upload_uri(dataset: Dataset, entity_path: Union[Path, str]) -> str:
    """Return the remote storage path that a dataset file would be located.

    Args:
        dataset(Dataset): Dataset with a backend storage.
        entity_path(Union[Path, str]): Dataset file's path (entity path); it is relative to the project's root.

    Returns:
        str: URI within remote storage.
    """
    assert dataset.storage, "Cannot get URI for datasets with no backend storage"

    base = dataset.storage.rstrip("/")
    path_within_dataset = get_dataset_file_path_within_dataset(dataset=dataset, entity_path=entity_path)

    return f"{base}/{path_within_dataset}"


def move_files_to_dataset(dataset: Dataset, files: List[DatasetAddMetadata]):
    """Copy/Move files into a dataset's directory."""

    def move_file(file: DatasetAddMetadata, storage: Optional[IStorage]):
        if not file.has_action:
            return

        if file.action in (
            DatasetAddAction.COPY,
            DatasetAddAction.MOVE,
            DatasetAddAction.SYMLINK,
            DatasetAddAction.DOWNLOAD,
        ):
            # NOTE: Remove existing file if any; required as a safety-net to avoid corrupting external files
            delete_dataset_file(file.destination, follow_symlinks=True)
            file.destination.parent.mkdir(parents=True, exist_ok=True)

        track_in_lfs = True

        # NOTE: If file is in a sub-directory of a dataset's remote storage URI, only update the metadata
        if file.remote_storage:
            if dataset.storage and is_uri_subfolder(dataset.storage, file.url):
                file.action = DatasetAddAction.METADATA_ONLY
            else:
                file.action = DatasetAddAction.DOWNLOAD

        if file.action == DatasetAddAction.COPY:
            shutil.copy(file.source, file.destination)
        elif file.action == DatasetAddAction.MOVE:
            shutil.move(file.source, file.destination, copy_function=shutil.copy)  # type: ignore
        elif file.action == DatasetAddAction.SYMLINK:
            create_external_file(target=file.source, path=file.destination)
            # NOTE: Don't track symlinks to external files in LFS
            track_in_lfs = False
        elif file.action == DatasetAddAction.DOWNLOAD:
            assert file.provider, f"Storage provider isn't set for {file} with DOWNLOAD action"
            storage = file.provider.get_storage()
            storage.download(file.url, file.destination)
        elif file.metadata_only:
            # NOTE: Nothing to do when adding file to a dataset with a parent remote storage
            pass
        else:
            raise errors.OperationError(f"Invalid action {file.action}")

        if track_in_lfs and not dataset.storage:
            track_paths_in_storage(file.destination)

        # NOTE: We always copy the files to the dataset's data dir. If dataset has a storage backend, we also upload the
        # file to the remote storage.
        if storage:
            if file.metadata_only:
                assert file.based_on, f"wasBasedOn isn't set for {file} with METADATA_ONLY action"
                file_uri = file.based_on.url
                md5_hash = file.based_on.checksum
            else:
                file_uri = get_upload_uri(dataset=dataset, entity_path=file.entity_path)
                storage.upload(source=file.destination, uri=file_uri)
                md5_hash = hash_file(file.destination, hash_type="md5") or ""

            file.based_on = RemoteEntity(url=file_uri, path=file.entity_path, checksum=md5_hash)

    dataset_storage = None
    if dataset.storage:
        provider = ProviderFactory.get_storage_provider(uri=dataset.storage)
        dataset_storage = provider.get_storage()

    for dataset_file in files:
        # TODO: Parallelize copy/download/upload
        move_file(file=dataset_file, storage=dataset_storage)


def add_files_to_repository(dataset: Dataset, files: List[DatasetAddMetadata]):
    """Track files in project's repository."""
    # NOTE: There is nothing to track for remote storages
    if dataset.storage:
        communication.info("Nothing to add to the project for datasets with a storage backend")
        return

    # NOTE: Don't commit files that will be uploaded to a remote storage
    paths_to_commit = [f.get_absolute_commit_path(project_context.path) for f in files]

    repository = project_context.repository

    # Force-add to include possible ignored files
    if len(paths_to_commit) > 0:
        repository.add(*paths_to_commit, project_context.pointers_path, force=True)

    n_staged_changes = len(repository.staged_changes)
    if n_staged_changes == 0:
        communication.warn("No new file was added to project")


def update_dataset_metadata(dataset: Dataset, files: List[DatasetAddMetadata], clear_files_before: bool):
    """Add newly-added files to the dataset's metadata."""
    dataset_files = []
    for file in files:
        dataset_file = DatasetFile.from_path(path=file.entity_path, source=file.url, based_on=file.based_on)
        dataset_files.append(dataset_file)

    if clear_files_before:
        dataset.clear_files()

    dataset.add_or_update_files(dataset_files)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Docker based interactive session provider."""

from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, cast
from uuid import uuid4

import docker

from renku.core import errors
from renku.core.config import get_value
from renku.core.plugin import hookimpl
from renku.core.util import communication
from renku.domain_model.project_context import project_context
from renku.domain_model.session import ISessionProvider, Session


class DockerSessionProvider(ISessionProvider):
    """A docker based interactive session provider."""

    JUPYTER_PORT = 8888

    def __init__(self):
        self._docker_client = None

    def docker_client(self) -> docker.client.DockerClient:
        """Get the docker client.

        Note:
            This is not a @property, even though it should be, because ``pluggy``
            will call it in that case in unrelated parts of the code that will
        Raises:
            errors.DockerError: Exception when docker is not available.
        Returns:
            The docker client.
        """
        if self._docker_client is None:
            try:
                self._docker_client = docker.from_env()
            except docker.errors.DockerException as e:
                raise errors.DockerError(f"Error while initializing docker client: {str(e)}")
        return self._docker_client

    @staticmethod
    def _get_jupyter_urls(ports: Dict[str, Any], auth_token: str, jupyter_port: int = 8888) -> Iterable[str]:
        port_key = f"{jupyter_port}/tcp"
        if port_key not in ports:
            return list()
        return map(lambda x: f'http://{x["HostIp"]}:{x["HostPort"]}/?token={auth_token}', ports[port_key])

    def _get_docker_containers(self, project_name: str) -> List[docker.models.containers.Container]:
        return self.docker_client().containers.list(filters={"label": f"renku_project={project_name}"})

    def get_name(self) -> str:
        """Return session provider's name."""
        return "docker"

    def build_image(self, image_descriptor: Path, image_name: str, config: Optional[Dict[str, Any]]):
        """Builds the container image."""
        self.docker_client().images.build(path=str(image_descriptor), tag=image_name)

    def find_image(self, image_name: str, config: Optional[Dict[str, Any]]) -> bool:
        """Find the given container image."""
        try:
            self.docker_client().images.get(image_name)
        except docker.errors.ImageNotFound:
            try:
                with communication.busy(msg=f"Pulling image from remote {image_name}"):
                    self.docker_client().images.pull(image_name)
            except docker.errors.NotFound:
                return False
            else:
                return True
        else:
            return True

    @hookimpl
    def session_provider(self) -> ISessionProvider:
        """Supported session provider.

        Returns:
            a reference to ``self``.
        """
        return self

    def session_list(self, project_name: str, config: Optional[Dict[str, Any]]) -> List[Session]:
        """Lists all the sessions currently running by the given session provider.

        Returns:
            list: a list of sessions.
        """

        return list(
            map(
                lambda x: Session(
                    id=cast(str, x.short_id),
                    status=x.status,
                    url=next(iter(DockerSessionProvider()._get_jupyter_urls(x.ports, x.labels["jupyter_token"]))),
                ),
                self._get_docker_containers(project_name),
            )
        )

    def session_start(
        self,
        image_name: str,
        project_name: str,
        config: Optional[Dict[str, Any]],
        cpu_request: Optional[float] = None,
        mem_request: Optional[str] = None,
        disk_request: Optional[str] = None,
        gpu_request: Optional[str] = None,
    ) -> str:
        """Creates an interactive session.

        Returns:
            str: a unique id for the created interactive session.
        """

        try:
            docker_is_running = self.docker_client().ping()
            if not docker_is_running:
                raise errors.DockerError(
                    "Could not communicate with the docker instance. Please make sure it is running!"
                )

            auth_token = uuid4().hex
            default_url = get_value("interactive", "default_url")

            # resource requests
            resource_requests: Dict[str, Any] = dict()
            if cpu_request:
                # based on the docker go cli: func ParseCPUs
                resource_requests["nano_cpus"] = int(cpu_request * 10**9)

            if mem_request:
                resource_requests["mem_limit"] = mem_request

            if disk_request:
                resource_requests["storage_opt"] = {"size": disk_request}

            if gpu_request:
                if gpu_request == "all":
                    resource_requests["device_requests"] = [
                        docker.types.DeviceRequest(count=-1, capabilities=[["compute", "utility"]])
                    ]
                else:
                    resource_requests["device_requests"] = [
                        docker.types.DeviceRequest(count=[gpu_request], capabilities=[["compute", "utility"]])
                    ]

            # NOTE: set git user
            image_data = self.docker_client().api.inspect_image(image_name)
            working_dir = image_data.get("Config", {}).get("WorkingDir", None)

            if working_dir is None:
                working_dir = "/home/jovyan"

            work_dir = Path(working_dir) / "work" / project_name.split("/")[-1]

            volumes = [f"{str(project_context.path.resolve())}:{work_dir}"]

            user = project_context.repository.get_user()
            environment = {
                "GIT_AUTHOR_NAME": user.name,
                "GIT_AUTHOR_EMAIL": user.email,
                "GIT_COMMITTER_EMAIL": user.email,
                "EMAIL": user.email,
            }

            container = self.docker_client().containers.run(
                image_name,
                f'jupyter notebook --NotebookApp.ip="0.0.0.0" --NotebookApp.port={DockerSessionProvider.JUPYTER_PORT}'
                f' --NotebookApp.token="{auth_token}" --NotebookApp.default_url="{default_url}"'
                f" --NotebookApp.notebook_dir={work_dir}",
                detach=True,
                labels={"renku_project": project_name, "jupyter_token": auth_token},
                ports={f"{DockerSessionProvider.JUPYTER_PORT}/tcp": None},
                remove=True,
                environment=environment,
                volumes=volumes,
                working_dir=str(work_dir),
                **resource_requests,
            )

            if not container.ports:
                container.reload()

            jupyter_urls = DockerSessionProvider._get_jupyter_urls(
                container.ports, auth_token, jupyter_port=DockerSessionProvider.JUPYTER_PORT
            )
            message = f"The session for '{image_name}' has been successfully started. It is available at:\n\t"
            message += "\n\t".join(jupyter_urls)
            return message
        except (docker.errors.APIError, docker.errors.BuildError) as error:
            raise errors.DockerError(str(error))

    def session_stop(self, project_name: str, session_name: Optional[str], stop_all: bool) -> bool:
        """Stops all or a given interactive session."""
        try:
            docker_containers = (
                self._get_docker_containers(project_name)
                if stop_all
                else self.docker_client().containers.list(filters={"id": session_name})
            )

            if len(docker_containers) == 0:
                return False

            [c.stop() for c in docker_containers]
            return True
        except docker.errors.APIError as error:
            raise errors.DockerError(error.msg)

    def session_url(self, session_name: str) -> Optional[str]:
        """Get the URL of the interactive session."""
        for c in self.docker_client().containers.list():
            if c.short_id == session_name and f"{DockerSessionProvider.JUPYTER_PORT}/tcp" in c.ports:
                host = c.ports[f"{DockerSessionProvider.JUPYTER_PORT}/tcp"][0]
                return f'http://{host["HostIp"]}:{host["HostPort"]}/?token={c.labels["jupyter_token"]}'
        return None



# -*- coding: utf-8 -*-
#
# Copyright 2017-2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resolution of ``Workflow`` execution values precedence."""

from abc import ABC, abstractmethod
from itertools import chain
from typing import Any, Dict, Optional, Set

from renku.core import errors
from renku.core.util.template_vars import TemplateVariableFormatter
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.parameter import ParameterMapping
from renku.domain_model.workflow.plan import AbstractPlan, Plan


class ValueResolver(ABC):
    """Value resolution class for an ``AbstractPlan``."""

    def __init__(self, plan: AbstractPlan, values: Optional[Dict[str, Any]]):
        self._values = values
        self.missing_parameters: Set[str] = set()
        self._plan = plan
        self._template_engine = TemplateVariableFormatter()

    @abstractmethod
    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a potentially nested workflow.

        Returns:
            AbstractPlan: The ``AbstractPlan`` with the user provided values set.
        """
        pass

    @staticmethod
    def get(plan: AbstractPlan, values: Dict[str, Any]) -> "ValueResolver":
        """Factory method to obtain the specific ValueResolver for a workflow.

        Args:
            plan(AbstractPlan): a workflow.
            values(Dict[str, Any]): user defined dictionary of runtime values for the provided workflow.

        Returns:
            "ValueResolver": A ValueResolver object.
        """
        return PlanValueResolver(plan, values) if isinstance(plan, Plan) else CompositePlanValueResolver(plan, values)


class PlanValueResolver(ValueResolver):
    """Value resolution class for a ``Plan``.

    Applies values and default_values to a workflow.
    """

    def __init__(self, plan: Plan, values: Dict[str, Any]):
        super(PlanValueResolver, self).__init__(plan, values)

    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a ``Plan``.

        Returns:
            A Plan with values applied.
        """
        if self._values is None:
            return self._plan

        values_keys = set(self._values.keys())
        for param in chain(self._plan.inputs, self._plan.parameters, self._plan.outputs):
            if param.name in self._values:
                param.actual_value = self._values[param.name]
                values_keys.discard(param.name)

        # NOTE: we need 2-pass the plan parameters as values can be overridden
        # that should be reflected in the params_map
        params_map = TemplateVariableFormatter.to_map(chain(self._plan.inputs, self._plan.parameters))
        for param in chain(self._plan.inputs, self._plan.parameters, self._plan.outputs):
            if isinstance(param.actual_value, str):
                actual_value_set_before = getattr(param, "_v_actual_value_set", False)
                param.actual_value = self._template_engine.apply(param.actual_value, params_map)
                param._v_actual_value_set = actual_value_set_before

        self.missing_parameters = values_keys

        return self._plan


class CompositePlanValueResolver(ValueResolver):
    """Value resolution class for a ``CompositePlan``.

    Applies values and default_values to a nested workflow.

    Order of precedence is as follows (from lowest to highest):
    - Default value on a parameter
    - Default value on a mapping to the parameter
    - Value passed to a mapping to the parameter
    - Value passed to the parameter
    - Value propagated to a parameter from the source of a ParameterLink
    """

    def __init__(self, plan: CompositePlan, values: Optional[Dict[str, Any]] = None):
        super(CompositePlanValueResolver, self).__init__(plan, values)

    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a ``CompositePlan``.

        Returns:
            A ``CompositePlan`` with values applied.
        """

        if self._values:
            self._apply_parameters_values()

            for name, step in filter(lambda x: isinstance(x[1], dict), self._values.items()):
                child_workflow = next((w for w in self._plan.plans if w.name == name), None)
                if not child_workflow:
                    raise errors.ChildWorkflowNotFoundError(name, self._plan.name)

                rv = ValueResolver.get(child_workflow, step)
                _ = rv.apply()
                self.missing_parameters.update({f"{name}.{mp}" for mp in rv.missing_parameters})

        # apply defaults
        for mapping in self._plan.mappings:
            self._apply_parameter_defaults(mapping)

        apply_parameter_links(self._plan)

        return self._plan

    def _apply_parameter_defaults(self, mapping: ParameterMapping) -> None:
        """Apply default values to a mapping and contained params if they're not set already.

        Args:
            mapping(ParameterMapping): The mapping to apply values to.
        """

        if not mapping.actual_value_set and mapping.default_value:
            mapping.actual_value = mapping.default_value

            for mapped_to in mapping.mapped_parameters:
                if isinstance(mapped_to, ParameterMapping):
                    self._apply_parameter_defaults(mapped_to)
                else:
                    if not mapped_to.actual_value_set:
                        mapped_to.actual_value = mapping.default_value

    def _apply_parameters_values(self) -> None:
        """Apply values to mappings of a CompositePlan."""
        if self._values is None:
            return
        for k, v in filter(lambda x: not isinstance(x[1], dict), self._values.items()):
            mapping = next((m for m in self._plan.mappings if m.name == k), None)

            if not mapping:
                self.missing_parameters.add(k)
                continue

            mapping.actual_value = v


def apply_parameter_links(workflow: CompositePlan) -> None:
    """Apply values from parameter links.

    Args:
        workflow(CompositePlan): The workflow whose links values should be applied on.
    """
    for link in workflow.links:
        for sink in link.sinks:
            sink.actual_value = link.source.actual_value

    for plan in workflow.plans:
        if isinstance(plan, CompositePlan):
            apply_parameter_links(plan)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['renku',
 'renku.command',
 'renku.command.checks',
 'renku.command.command_builder',
 'renku.command.format',
 'renku.command.schema',
 'renku.command.view_model',
 'renku.core',
 'renku.core.dataset',
 'renku.core.dataset.providers',
 'renku.core.interface',
 'renku.core.migration',
 'renku.core.migration.models',
 'renku.core.migration.utils',
 'renku.core.plugin',
 'renku.core.plugin.implementations',
 'renku.core.session',
 'renku.core.template',
 'renku.core.util',
 'renku.core.workflow',
 'renku.core.workflow.converters',
 'renku.core.workflow.model',
 'renku.core.workflow.parser',
 'renku.core.workflow.providers',
 'renku.data',
 'renku.domain_model',
 'renku.domain_model.provenance',
 'renku.domain_model.workflow',
 'renku.domain_model.workflow.converters',
 'renku.infrastructure',
 'renku.infrastructure.gateway',
 'renku.infrastructure.storage',
 'renku.ui',
 'renku.ui.api',
 'renku.ui.api.graph',
 'renku.ui.api.models',
 'renku.ui.cli',
 'renku.ui.cli.utils',
 'renku.ui.service',
 'renku.ui.service.cache',
 'renku.ui.service.cache.models',
 'renku.ui.service.cache.serializers',
 'renku.ui.service.controllers',
 'renku.ui.service.controllers.api',
 'renku.ui.service.controllers.utils',
 'renku.ui.service.gateways',
 'renku.ui.service.interfaces',
 'renku.ui.service.jobs',
 'renku.ui.service.serializers',
 'renku.ui.service.serializers.v0_9',
 'renku.ui.service.serializers.v1_0',
 'renku.ui.service.utils',
 'renku.ui.service.views',
 'renku.ui.service.views.v0_9',
 'renku.ui.service.views.v1_0']

package_data = \
{'': ['*'],
 'renku': ['templates/*',
           'templates/R-minimal/*',
           'templates/R-minimal/.renku/*',
           'templates/R-minimal/data/*',
           'templates/R-minimal/notebooks/*',
           'templates/bioc-minimal/*',
           'templates/bioc-minimal/.renku/*',
           'templates/bioc-minimal/data/*',
           'templates/bioc-minimal/notebooks/*',
           'templates/julia-minimal/*',
           'templates/julia-minimal/.renku/*',
           'templates/julia-minimal/data/*',
           'templates/julia-minimal/notebooks/*',
           'templates/minimal/*',
           'templates/minimal/.renku/*',
           'templates/python-minimal/*',
           'templates/python-minimal/.renku/*',
           'templates/python-minimal/data/*',
           'templates/python-minimal/notebooks/*']}

install_requires = \
['appdirs>=1.4.3,<=1.4.4',
 'attrs>=21.1.0,<22.2.0',
 'bashlex>=0.16,<0.17',
 'calamus>=0.3.13,<0.5',
 'click-option-group>=0.5.2,<0.6.0',
 'click-plugins==1.1.1',
 'click>=8.0,<8.1.4',
 'cryptography>=3.4.1,<38.0.0',
 'cwl-utils>=0.12,<0.18',
 'cwltool==3.1.20220628170238',
 'deepdiff>=5.8.0,<6.0.0',
 'deepmerge==1.0.1',
 'docker>=3.7.2,<6',
 'filelock>=3.3.0,<3.8.1',
 'gitpython==3.1.27',
 'grandalf==0.7',
 'humanize>=3.0.0,<4.1.0',
 'inject>=4.3.0,<4.4.0',
 'jinja2>=2.11.3,<3.1.3',
 'networkx>=2.6.0,<2.7',
 'numpy>=1.20.0,<1.22.0',
 'packaging>=21.3,<22.0',
 'pathspec>=0.8.0,<1.0.0',
 'patool==1.12',
 'pluggy==1.0.0',
 'portalocker>=2.2.1,<2.5',
 'psutil>=5.4.7,<5.9.2',
 'pydantic==1.10.2',
 'pyjwt>=2.1.0,<2.5.0',
 'pyld==2.0.3',
 'pyopenssl>=19.0.0,<22.1.0',
 'pyshacl>=0.17.2,<0.18.2',
 'python-dateutil>=2.6.1,<2.8.3',
 'python-editor==1.0.4',
 'python-gitlab>=2.10.1,<3.8.2',
 'pyyaml>=5.4,<6.1.0',
 'rdflib>=6.0.0,<7.0',
 'requests>=2.23.0,<2.28.2',
 'rich>=9.3.0,<12.6.0',
 'shellingham==1.4.0',
 'tabulate>=0.7.7,<0.8.11',
 'tqdm>=4.48.1,<4.62.4',
 'werkzeug>=1.0.0,<2.2.3',
 'yagup>=0.1.1',
 'yaspin==2.1.0',
 'zc.relation>=1.1,<1.2',
 'zodb==5.6.0',
 'zstandard>=0.16.0,<0.18.0']

extras_require = \
{':extra == "all"': ['toil==5.7.1'],
 ':python_full_version < "3.8.0"': ['importlib-metadata>=4.8.1,<4.13.0',
                                    'typing-extensions>=4.3.0,<5.0.0'],
 ':python_full_version < "3.9.0"': ['importlib-resources>=5.4.0,<5.10.0'],
 'all': ['apispec>=4.0.0,<5.3.0',
         'apispec-webframeworks>=0.5.2,<0.6',
         'black==22.6.0',
         'circus==0.17.1',
         'coverage[toml]>=4.5.3,<6.5',
         'fakeredis[lua]>=1.4.1,<1.7.2',
         'flakehell>=0.9.0,<1.0',
         'flaky==3.7.0',
         'flask==2.1.1',
         'gunicorn',
         'isort>=5.3.2,<5.10.2',
         'marshmallow>=3.18.0',
         'marshmallow-oneofschema>=3.0.1,<4.0.0',
         'mypy>=0.942,<1.0',
         'pexpect>=4.8.0,<4.9.0',
         'pillow>=9.0.0,<9.4',
         'plantweb>=1.2.1,<1.3.0',
         'poetry-dynamic-versioning>=0.14,<0.18',
         'pydocstyle>=4.0.1,<6.1.2',
         'pyte>=0.8.0,<0.9.0',
         'pytest>=4.0.0,<7.1.4',
         'pytest-black>=0.3.10,<0.3.13',
         'pytest-cache==1.0',
         'pytest-cov>=2.5.1,<3.1.0',
         'pytest-flake8>=1.0.6,<1.1.1',
         'pytest-mock>=3.2.0,<3.9.0',
         'pytest-pep8==1.0.6',
         'pytest-recording==0.12.1',
         'pytest-timeout==2.1.0',
         'pytest-xdist>=1.34.0,<2.6.0',
         'python-dotenv>=0.19.0,<0.21.0',
         'redis>=3.5.3,<4.2.0',
         'renku-sphinx-theme>=0.2.0',
         'responses>=0.22.0,<0.23.0',
         'rq==1.11.0',
         'rq-scheduler==0.11.0',
         'sentry-sdk[flask]>=1.5.11,<1.5.12',
         'sphinxcontrib-spelling>=7.0.0,<8.0.0',
         'sphinx-click>=4.3.0,<5.0.0',
         'sphinx-rtd-theme>=0.5.0,<1.1',
         'sphinx-tabs==3.2.0',
         'types-python-dateutil>=2.8.10,<3.0.0',
         'types-PyYAML>=5.4,<6.1.0',
         'types-redis>=3.5.3,<4.1.0',
         'types-requests>=2.23.0,<2.27.2',
         'types-tabulate>=0.7.7,<0.8.10',
         'walrus>=0.8.2,<0.10.0'],
 'docs': ['plantweb>=1.2.1,<1.3.0',
          'renku-sphinx-theme>=0.2.0',
          'sphinxcontrib-spelling>=7.0.0,<8.0.0',
          'sphinx-click>=4.3.0,<5.0.0',
          'sphinx-rtd-theme>=0.5.0,<1.1',
          'sphinx-tabs==3.2.0'],
 'service': ['apispec>=4.0.0,<5.3.0',
             'apispec-webframeworks>=0.5.2,<0.6',
             'circus==0.17.1',
             'flask==2.1.1',
             'gunicorn',
             'marshmallow>=3.18.0',
             'marshmallow-oneofschema>=3.0.1,<4.0.0',
             'pillow>=9.0.0,<9.4',
             'python-dotenv>=0.19.0,<0.21.0',
             'redis>=3.5.3,<4.2.0',
             'rq==1.11.0',
             'rq-scheduler==0.11.0',
             'sentry-sdk[flask]>=1.5.11,<1.5.12',
             'walrus>=0.8.2,<0.10.0'],
 'tests': ['black==22.6.0',
           'coverage[toml]>=4.5.3,<6.5',
           'fakeredis[lua]>=1.4.1,<1.7.2',
           'flakehell>=0.9.0,<1.0',
           'flaky==3.7.0',
           'isort>=5.3.2,<5.10.2',
           'mypy>=0.942,<1.0',
           'pexpect>=4.8.0,<4.9.0',
           'poetry-dynamic-versioning>=0.14,<0.18',
           'pydocstyle>=4.0.1,<6.1.2',
           'pyte>=0.8.0,<0.9.0',
           'pytest>=4.0.0,<7.1.4',
           'pytest-black>=0.3.10,<0.3.13',
           'pytest-cache==1.0',
           'pytest-cov>=2.5.1,<3.1.0',
           'pytest-flake8>=1.0.6,<1.1.1',
           'pytest-mock>=3.2.0,<3.9.0',
           'pytest-pep8==1.0.6',
           'pytest-recording==0.12.1',
           'pytest-timeout==2.1.0',
           'pytest-xdist>=1.34.0,<2.6.0',
           'responses>=0.22.0,<0.23.0',
           'types-python-dateutil>=2.8.10,<3.0.0',
           'types-PyYAML>=5.4,<6.1.0',
           'types-redis>=3.5.3,<4.1.0',
           'types-requests>=2.23.0,<2.27.2',
           'types-tabulate>=0.7.7,<0.8.10']}

entry_points = \
{'console_scripts': ['renku = renku.ui.cli:cli']}

setup_kwargs = {
    'name': 'renku',
    'version': '2.0.0',
    'description': 'Python SDK and CLI for the Renku platform.',
    'long_description': '..\n    Copyright 2017-2022 - Swiss Data Science Center (SDSC)\n    A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and\n    Eidgenössische Technische Hochschule Zürich (ETHZ).\n\n    Licensed under the Apache License, Version 2.0 (the "License");\n    you may not use this file except in compliance with the License.\n    You may obtain a copy of the License at\n\n        http://www.apache.org/licenses/LICENSE-2.0\n\n    Unless required by applicable law or agreed to in writing, software\n    distributed under the License is distributed on an "AS IS" BASIS,\n    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n    See the License for the specific language governing permissions and\n    limitations under the License.\n\n======================================\n Renku Python Library, CLI and Service\n======================================\n\n.. image:: https://github.com/SwissDataScienceCenter/renku-python/workflows/Test,%20Integration%20Tests%20and%20Deploy/badge.svg\n   :target: https://github.com/SwissDataScienceCenter/renku-python/actions?query=workflow%3A%22Test%2C+Integration+Tests+and+Deploy%22+branch%3Amaster\n\n.. image:: https://img.shields.io/coveralls/SwissDataScienceCenter/renku-python.svg\n   :target: https://coveralls.io/r/SwissDataScienceCenter/renku-python\n\n.. image:: https://img.shields.io/github/tag/SwissDataScienceCenter/renku-python.svg\n   :target: https://github.com/SwissDataScienceCenter/renku-python/releases\n\n.. image:: https://img.shields.io/pypi/dm/renku.svg\n   :target: https://pypi.python.org/pypi/renku\n\n.. image:: http://readthedocs.org/projects/renku-python/badge/?version=latest\n   :target: http://renku-python.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://img.shields.io/github/license/SwissDataScienceCenter/renku-python.svg\n   :target: https://github.com/SwissDataScienceCenter/renku-python/blob/master/LICENSE\n\nA Python library for the `Renku collaborative data science platform\n<https://github.com/SwissDataScienceCenter/renku>`_. It includes a CLI and SDK\nfor end-users as well as a service backend. It provides functionality for the\ncreation and management of projects and datasets, and simple utilities to\ncapture data provenance while performing analysis tasks.\n\n**NOTE**:\n   ``renku-python`` is the python library and core service for Renku - it *does\n   not* start the Renku platform itself - for that, refer to the Renku docs on\n   `running the platform\n   <https://renku.readthedocs.io/en/latest/developer/setup.html>`_.\n\n\nRenku for Users\n===============\n\nInstallation\n------------\n\n.. _installation-reference:\n\nRenku releases and development versions are available from `PyPI\n<https://pypi.org/project/renku/>`_. You can install it using any tool that\nknows how to handle PyPI packages. Our recommendation is to use `:code:pipx\n<https://github.com/pipxproject/pipx>`_.\n\n.. note::\n\n   We do not officially support Windows at this moment. The way Windows\n   handles paths and symlinks interferes with some Renku functionality.\n   We recommend using the Windows Subsystem for Linux (WSL) to use Renku\n   on Windows.\n\nPrerequisites\n~~~~~~~~~~~~~\n\nRenku depends on Git under the hood, so make sure that you have Git\n`installed on your system <https://git-scm.com/downloads>`_.\n\nRenku also offers support to store large files in `Git LFS\n<https://git-lfs.github.com/>`_, which is used by default and should be\ninstalled on your system. If you do not wish to use Git LFS, you can run\nRenku commands with the `-S` flag, as in `renku -S <command>`.  More\ninformation on Git LFS usage in renku can be found in the `Data in Renku\n<https://renku.readthedocs.io/en/latest/user/data.html>`_ section of the docs.\n\nRenku uses CWL to execute recorded workflows when calling `renku update`\nor `renku rerun`. CWL depends on NodeJs to execute the workflows, so installing\n`NodeJs <https://nodejs.org/en/download/package-manager/>`_ is required if\nyou want to use those features.\n\nFor development of the service, `Docker <https://docker.com>`_ is recommended.\n\n\n.. _pipx-before-reference:\n\n``pipx``\n~~~~~~~~\n.. _pipx-after-reference:\n\nFirst, `install pipx <https://github.com/pipxproject/pipx#install-pipx>`_\nand make sure that the ``$PATH`` is correctly configured.\n\n::\n\n    $ python3 -m pip install --user pipx\n    $ python3 -m pipx ensurepath\n\nOnce ``pipx`` is installed use following command to install ``renku``.\n\n::\n\n    $ pipx install renku\n    $ which renku\n    ~/.local/bin/renku\n\n\n``pipx`` installs Renku into its own virtual environment, making sure that it\ndoes not pollute any other packages or versions that you may have already\ninstalled.\n\n.. note::\n\n    If you install Renku as a dependency in a virtual environment and the\n    environment is active, your shell will default to the version installed\n    in the virtual environment, *not* the version installed by ``pipx``.\n\n\nTo install a development release:\n\n::\n\n    $ pipx install --pip-args pre renku\n\n\n.. _pip-before-reference:\n\n``pip``\n~~~~~~~\n.. _pip-after-reference:\n\n::\n\n    $ pip install renku\n\nThe latest development versions are available on PyPI or from the Git\nrepository:\n\n::\n\n    $ pip install --pre renku\n    # - OR -\n    $ pip install -e git+https://github.com/SwissDataScienceCenter/renku-python.git#egg=renku\n\nUse following installation steps based on your operating system and preferences\nif you would like to work with the command line interface and you do not need\nthe Python library to be importable.\n\n.. _windows-before-reference:\n\nWindows\n~~~~~~~\n.. _windows-after-reference:\n\n.. note::\n\n    We don\'t officially support Windows yet, but Renku works well in the Windows Subsystem for Linux (WSL).\n    As such, the following can be regarded as a best effort description on how to get started with Renku on Windows.\n\nRenku can be run using the Windows Subsystem for Linux (WSL). To install the WSL, please follow the\n`official instructions <https://docs.microsoft.com/en-us/windows/wsl/install-win10#manual-installation-steps>`__.\n\nWe recommend you use the Ubuntu 20.04 image in the WSL when you get to that step of the installation.\n\nOnce WSL is installed, launch the WSL terminal and install the packages required by Renku with:\n\n::\n\n    $ sudo apt-get update && sudo apt-get install git python3 python3-pip python3-venv pipx\n\nSince Ubuntu has an older version of git LFS installed by default which is known to have some bugs when cloning\nrepositories, we recommend you manually install the newest version by following\n`these instructions <https://github.com/git-lfs/git-lfs/wiki/Installation#debian-and-ubuntu>`__.\n\nOnce all the requirements are installed, you can install Renku normally by running:\n\n::\n\n    $ pipx install renku\n    $ pipx ensurepath\n\nAfter this, Renku is ready to use. You can access your Windows in the various mount points in\n``/mnt/`` and you can execute Windows executables (e.g. ``\\*.exe``) as usual directly from the\nWSL (so ``renku run myexecutable.exe`` will work as expected).\n\n.. _docker-before-reference:\n\nDocker\n~~~~~~\n.. _docker-after-reference:\n\nThe containerized version of the CLI can be launched using Docker command.\n\n::\n\n    $ docker run -it -v "$PWD":"$PWD" -w="$PWD" renku/renku-python renku\n\nIt makes sure your current directory is mounted to the same place in the\ncontainer.\n\n\nCLI Example\n-----------\n\nInitialize a Renku project:\n\n::\n\n    $ mkdir -p ~/temp/my-renku-project\n    $ cd ~/temp/my-renku-project\n    $ renku init\n\nCreate a dataset and add data to it:\n\n::\n\n    $ renku dataset create my-dataset\n    $ renku dataset add my-dataset https://raw.githubusercontent.com/SwissDataScienceCenter/renku-python/master/README.rst\n\nRun an analysis:\n\n::\n\n    $ renku run --name my-workflow -- wc < data/my-dataset/README.rst > wc_readme\n\nTrace the data provenance:\n\n::\n\n    $ renku workflow visualize wc_readme\n\nThese are the basics, but there is much more that Renku allows you to do with\nyour data analysis workflows. The full documentation will soon be available\nat: https://renku-python.readthedocs.io/\n\n\nRenku as a Service\n==================\n\nThis repository includes a ``renku-core`` RPC service written as a `Flask\n<https://flask.palletsprojects.com>`_ application that provides (almost) all of\nthe functionality of the Renku CLI. This is used to provide one of the backends\nfor the `RenkuLab <https://renkulab.io>`_ web UI. The service can be deployed in\nproduction as a Helm chart (see `helm-chart <./helm-chart/README.rst>`_.\n\n\nDeploying locally\n-----------------\n\nTo test the service functionality you can deploy it quickly and easily using\n``docker-compose up`` [docker-compose](https://pypi.org/project/docker-compose/).\nMake sure to make a copy of the ``renku/service/.env-example`` file and configure it\nto your needs. The setup here is to expose the service behind a traefik reverse proxy\nto mimic an actual production deployment. You can access the proxied endpoints at\n``http://localhost/api``. The service itself is exposed on port 8080 so its endpoints\nare available directly under ``http://localhost:8080``.\n\n\nAPI Documentation\n-----------------\n\nThe renku core service implements the API documentation as an OpenAPI 3.0.x spec.\nYou can retrieve the yaml of the specification itself with\n\n```\n$ renku service apispec\n```\n\nIf deploying the service locally with ``docker-compose`` you can find the swagger-UI\nunder ``localhost/api/swagger``. To send the proper authorization headers to the\nservice endpoints, click the ``Authorize`` button and enter a valid JWT token and\na gitlab token with read/write repository scopes. The JWT token can be obtained by\nlogging in to a renku instance with ``renku login`` and retrieving it from your local\nrenku configuration.\n\nIn a live deployment, the swagger documentation is available under ``https://<renku-endpoint>/swagger``.\nYou can authorize the API by first logging into renku normally, then going to the\nswagger page, clicking ``Authorize`` and picking the ``oidc (OAuth2, authorization_code)``\noption. Leave the ``client_id`` as ``swagger`` and the ``client_secret`` empty, select\nall scopes and click ``Authorize``. You should now be logged in and you can send\nrequests using the ``Try it out`` buttons on individual requests.\n\n\nDeveloping Renku\n================\n\nFor testing the functionality from source it is convenient to install ``renku``\nin editable mode using ``pipx``. Clone the repository and then do:\n\n::\n\n    $ pipx install \\\n        --editable \\\n        <path-to-renku-python>[all] \\\n        renku\n\nThis will install all the extras for testing and debugging.\n\nIf you already use `pyenv <https://github.com/pyenv/pyenv>`__ to manage different python versions,\nyou may be interested in installing `pyenv-virtualenv <https://github.com/pyenv/pyenv-virtualenv>`__ to\ncreate an ad-hoc virtual environment for developing renku.\n\nOnce you have created and activated a virtual environment for renku-python, you can use the usual\n`pip` commands to install the required dependencies.\n\n::\n\n    $ pip install -e .[all]  # use `.[all]` for zsh\n\n\nService\n-------\n\nDeveloping the service and testing its APIs can be done with ``docker compose`` (see\n"Deploying Locally" above).\n\nIf you have a full RenkuLab deployment at your disposal, you can\nuse `telepresence <https://www.telepresence.io/>`__ v1 to develop and debug locally.\nJust run the `start-telepresence.sh` script and follow the instructions.\nMind that the script doesn\'t work with telepresence v2.\n\n\nRunning tests\n-------------\n\nWe use `pytest <https://docs.pytest.org>`__ for running tests.\nYou can use our `run-tests.sh` script for running specific set of tests.\n\n::\n\n    $ ./run-tests.sh -h\n\nWe lint the files using `black <https://github.com/psf/black>`__ and\n`isort <https://github.com/PyCQA/isort>`__.\n\n\nUsing External Debuggers\n------------------------\n\nLocal Machine\n~~~~~~~~~~~~~\n\nTo run ``renku`` via e.g. the `Visual Studio Code debugger\n<https://code.visualstudio.com/docs/python/debugging>`_ you need run it via\nthe python executable in whatever virtual environment was used to install ``renku``. If there is a package\nneeded for the debugger, you need to inject it into the virtual environment first, e.g.:\n\n::\n\n    $ pipx inject renku ptvsd\n\n\nFinally, run ``renku`` via the debugger:\n\n::\n\n    $ ~/.local/pipx/venvs/renku/bin/python -m ptvsd --host localhost --wait -m renku.ui.cli <command>\n\n\nIf using Visual Studio Code, you may also want to set the ``Remote Attach`` configuration\n``PathMappings`` so that it will find your source code, e.g.\n\n::\n\n    {\n        "name": "Python: Remote Attach",\n        "type": "python",\n        "request": "attach",\n        "port": 5678,\n        "host": "localhost",\n        "pathMappings": [\n            {\n                "localRoot": "<path-to-renku-python-source-code>",\n                "remoteRoot": "<path-to-renku-python-source-code>"\n            }\n        ]\n    }\n',
    'author': 'Swiss Data Science Center',
    'author_email': 'contact@datascience.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/swissdatasciencecenter/renku-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks for external files."""

from renku.command.command_builder import inject
from renku.command.util import WARNING, red_text, yellow_text
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.domain_model.project_context import project_context


@inject.autoparams("dataset_gateway")
def check_missing_external_files(dataset_gateway: IDatasetGateway, **_):
    """Find external files that are missing.

    Args:
        dataset_gateway(IDatasetGateway): The injected dataset gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether no external files are missing and string of found problems.
    """
    missing = []

    for dataset in dataset_gateway.get_all_active_datasets():
        for file_ in dataset.files:
            if file_.is_external:
                target = (project_context.path / file_.entity.path).resolve()
                if not target.exists():
                    missing.append((file_.entity.path, str(target)))

    if not missing:
        return True, None

    missing_str = "\n\t".join(f"{yellow_text(path)} -> {red_text(target)}" for path, target in missing)
    problems = (
        f"\n{WARNING}There are missing external files.\n  (make sure that external paths are accessible)"
        f"\n\n\t{missing_str}\n"
    )
    return False, problems



# -*- coding: utf-8 -*-
#
# Copyright 2018-2022- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interactive session business logic."""

import webbrowser
from itertools import chain
from typing import List, Optional

from pydantic import validate_arguments

from renku.core import errors
from renku.core.config import get_value
from renku.core.plugin.session import get_supported_session_providers
from renku.core.session.utils import get_image_repository_host, get_renku_project_name
from renku.core.util import communication
from renku.core.util.os import safe_read_yaml
from renku.domain_model.session import ISessionProvider, Session


def _safe_get_provider(provider: str) -> ISessionProvider:
    try:
        return next(p for p in get_supported_session_providers() if p.get_name() == provider)
    except StopIteration:
        raise errors.ParameterError(f"Session provider '{provider}' is not available!")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_list(config_path: Optional[str], provider: Optional[str] = None):
    """List interactive sessions."""

    def list_sessions(session_provider: ISessionProvider) -> List[Session]:
        try:
            return session_provider.session_list(config=config, project_name=project_name)
        except errors.RenkulabSessionGetUrlError:
            if provider:
                raise
            return []

    project_name = get_renku_project_name()
    config = safe_read_yaml(config_path) if config_path else dict()

    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()

    return list(chain(*map(list_sessions, providers)))


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_start(
    provider: str,
    config_path: Optional[str],
    image_name: Optional[str] = None,
    cpu_request: Optional[float] = None,
    mem_request: Optional[str] = None,
    disk_request: Optional[str] = None,
    gpu_request: Optional[str] = None,
):
    """Start interactive session."""
    from renku.domain_model.project_context import project_context

    pinned_image = get_value("interactive", "image")
    if pinned_image and image_name is None:
        image_name = pinned_image

    provider_api = _safe_get_provider(provider)
    config = safe_read_yaml(config_path) if config_path else dict()

    provider_api.pre_start_checks()

    project_name = get_renku_project_name()
    if image_name is None:
        tag = project_context.repository.head.commit.hexsha[:7]
        repo_host = get_image_repository_host()
        image_name = f"{project_name}:{tag}"
        if repo_host:
            image_name = f"{repo_host}/{image_name}"

        if not provider_api.find_image(image_name, config):
            communication.confirm(
                f"The container image '{image_name}' does not exists. Would you like to build it?",
                abort=True,
            )
            with communication.busy(msg=f"Building image {image_name}"):
                _ = provider_api.build_image(project_context.docker_path.parent, image_name, config)
            communication.echo(f"Image {image_name} built successfully.")
    else:
        if not provider_api.find_image(image_name, config):
            raise errors.ParameterError(f"Cannot find the provided container image '{image_name}'!")

    # set resource settings
    cpu_limit = cpu_request or get_value("interactive", "cpu_request")

    if cpu_limit is not None:
        try:
            cpu_limit = float(cpu_limit)
        except ValueError:
            raise errors.SessionStartError(f"Invalid value for cpu_request (must be float): {cpu_limit}")

    disk_limit = disk_request or get_value("interactive", "disk_request")
    mem_limit = mem_request or get_value("interactive", "mem_request")
    gpu = gpu_request or get_value("interactive", "gpu_request")

    with communication.busy(msg="Waiting for session to start..."):
        session_name = provider_api.session_start(
            config=config,
            project_name=project_name,
            image_name=image_name,
            cpu_request=cpu_limit,
            mem_request=mem_limit,
            disk_request=disk_limit,
            gpu_request=gpu,
        )
    communication.echo(msg=f"Session {session_name} successfully started")
    return session_name


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_stop(session_name: Optional[str], stop_all: bool = False, provider: Optional[str] = None):
    """Stop interactive session."""

    def stop_sessions(session_provider: ISessionProvider) -> bool:
        try:
            return session_provider.session_stop(
                project_name=project_name, session_name=session_name, stop_all=stop_all
            )
        except errors.RenkulabSessionGetUrlError:
            if provider:
                raise
            return False

    session_detail = "all sessions" if stop_all else f"session {session_name}"
    project_name = get_renku_project_name()

    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()

    with communication.busy(msg=f"Waiting for {session_detail} to stop..."):
        is_stopped = any(map(stop_sessions, providers))

    if not is_stopped:
        if not session_name:
            raise errors.ParameterError("There are no running sessions.")
        raise errors.ParameterError(f"Could not find '{session_name}' among the running sessions.")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_open(session_name: str, provider: Optional[str] = None):
    """Open interactive session in the browser."""

    def open_sessions(session_provider: ISessionProvider) -> Optional[str]:
        try:
            return session_provider.session_url(session_name=session_name)
        except errors.RenkulabSessionGetUrlError:
            if provider:
                raise
            return None

    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()

    url = next(filter(lambda u: u is not None, map(open_sessions, providers)), None)

    if url is None:
        raise errors.ParameterError(f"Could not find '{session_name}' among the running sessions.")
    webbrowser.open(url)
